import decimal
import typing as ty
from decimal import InvalidOperation

__all__ = (
    "Decimal",
    "InvalidOperation",
)


class Decimal(decimal.Decimal):
    @classmethod
    def __new__(cls, *args, **kwargs) -> ty.Any:
        try:
            return super().__new__(*args, **kwargs)
        except InvalidOperation as e:
            raise InvalidOperation(f"{e} (check {args=!s})")



from .functions import text_to_decimal, text_to_label
from .sessions import (
    DEFAULT_REQUEST_TIMEOUT,
    DEFAULT_RETRY_ALLOWED_METHODS,
    DEFAULT_RETRY_BACKOFF_FACTOR,
    DEFAULT_RETRY_TOTAL,
    Session,
    get_session,
    make_request,
)



import json
import logging
import typing as ty
from urllib.parse import urljoin

from lxml import html
from lxml.etree import ParserError
from requests import Response, Session
from requests.adapters import HTTPAdapter
from requests.exceptions import ConnectionError, HTTPError
from urllib3 import Retry
from user_agent import generate_user_agent  # type: ignore[import-untyped]

from finavis.exceptions import (
    RequestDocumentIsEmptyException,
    RequestMaxRetryException,
    RequestUnhandledException,
    TickerNotFoundException,
)

logger = logging.getLogger(__name__)

DEFAULT_RETRY_TOTAL: int = 0
DEFAULT_RETRY_BACKOFF_FACTOR: int = 1800
DEFAULT_RETRY_ALLOWED_METHODS: ty.Sequence[str] = ("POST", "GET")
DEFAULT_REQUEST_TIMEOUT: int = 2


def get_session(
    retry_total: int = DEFAULT_RETRY_TOTAL,
    retry_backoff_factor: int = DEFAULT_RETRY_BACKOFF_FACTOR,
    retry_allowed_methods: ty.Sequence[str] = DEFAULT_RETRY_ALLOWED_METHODS,
    headers: ty.Optional[ty.Mapping] = None,
    proxy_url: ty.Optional[str] = None,
) -> Session:
    """Custom session object w/ some settings"""

    session: Session = Session()

    if headers is not None:
        session.headers.update(headers)

    session.headers.update({"User-Agent": generate_user_agent()})

    if proxy_url is not None:
        session.proxies = {"https": proxy_url, "http": proxy_url}

    if retry_total is not None:
        retries: Retry = Retry(
            total=retry_total,
            backoff_factor=retry_backoff_factor / 1000,
            allowed_methods=list(retry_allowed_methods),
        )
        session.mount("https://", HTTPAdapter(max_retries=retries))

    return session


def make_request(
    path: str,
    query_params: ty.Optional[ty.Dict[str, str]] = None,
) -> html.HtmlElement:
    """Make request to some URL"""

    params: ty.Dict[str, ty.Any] = dict(
        method="GET",
        url=urljoin("https://finviz.com/", path),
        params=query_params or {},
        allow_redirects=False,
        timeout=3,
    )

    logger.debug(
        f"make_request = {params['method']} {params['url']} "
        f"query_params={json.dumps(params['params'])}"
    )

    session: Session = get_session()

    try:
        response: Response = session.request(**params)
        response.raise_for_status()
    except (HTTPError, ConnectionError) as e:
        if isinstance(e, ConnectionError) and "Max retries exceeded" in str(e):
            raise RequestMaxRetryException(e)
        elif isinstance(e, HTTPError) and "404" in str(e) and "/quote.ashx" in str(e):
            raise TickerNotFoundException(e)
        else:
            raise RequestUnhandledException(e)
    else:
        logger.debug(f"make_request = DONE status_code={response.status_code}")

    try:
        return html.fromstring(html=response.text)
    except ParserError as e:
        if "document is empty" in str(e).lower():
            raise RequestDocumentIsEmptyException(e)
        else:
            raise RequestUnhandledException(e)



import typing as ty
from enum import StrEnum


class EnumWithValues(StrEnum):
    @classmethod
    def values(cls) -> ty.Tuple[str, ...]:
        return tuple(map(lambda c: str(c), cls))

    @classmethod
    def keys(cls) -> ty.Tuple[str, ...]:
        return tuple(map(lambda c: c.name, cls))  # type: ignore[attr-defined]


class Exchange(EnumWithValues):
    AMEX = "exch_amex"
    NASDAQ = "exch_nasd"
    NYSE = "exch_nyse"


class Index(EnumWithValues):
    SP500 = "idx_sp500"
    NASDAQ100 = "idx_ndx"
    DJIA = "idx_dji"
    RUSSELL2000 = "idx_rut"


class Order(EnumWithValues):
    TICKER_ASC = "ticker"
    TICKER_DESC = "-ticker"
    COMPANY_ASC = "company"
    COMPANY_DESC = "-company"
    SECTOR_ASC = "sector"
    SECTOR_DESC = "-sector"
    INDUSTRY_ASC = "industry"
    INDUSTRY_DESC = "-industry"
    COUNTRY_ASC = "country"
    COUNTRY_DESC = "-country"
    MARKETCAP_ASC = "marketcap"
    MARKETCAP_DESC = "-marketcap"
    PE_ASC = "pe"
    PE_DESC = "-pe"
    PRICE_ASC = "price"
    PRICE_DESC = "-price"
    CHANGE_ASC = "change"
    CHANGE_DESC = "-change"
    VOLUME_ASC = "volume"
    VOLUME_DESC = "-volume"


class Signal(EnumWithValues):
    TOP_GAINERS = "ta_topgainers"
    TOP_LOSERS = "ta_toplosers"
    NEW_HIGH = "ta_newhigh"
    NEW_LOW = "ta_newlow"
    MOST_VOLATILE = "ta_mostvolatile"
    MOST_ACTIVE = "ta_mostactive"
    UNUSUAL_VOLUME = "ta_unusualvolume"
    OVERBOUGHT = "ta_overbought"
    OVERSOLD = "ta_oversold"
    DOWNGRADES = "n_downgrades"
    UPGRADES = "n_upgrades"
    EARNINGS_BEFORE = "n_earningsbefore"
    EARNINGS_AFTER = "n_earningsafter"
    RECENT_INSIDER_BUYING = "it_latestbuys"
    RECENT_INSIDER_SELLING = "it_latestsales"
    MAJOR_NEWS = "n_majornews"
    HORIZONTAL_SR = "ta_p_horizontal"
    TL_RESISTANCE = "ta_p_tlresistance"
    TL_SUPPORT = "ta_p_tlsupport"
    WEDGE_UP = "ta_p_wedgeup"
    WEDGE_DOWN = "ta_p_wedgedown"
    TRIANGLE_ASC = "ta_p_wedgeresistance"
    TRIANGLE_DESC = "ta_p_wedgesupport"
    WEDGE = "ta_p_wedge"
    CHANNEL_UP = "ta_p_channelup"
    CHANNEL_DOWN = "ta_p_channeldown"
    CHANNEL = "ta_p_channel"
    DOUBLE_TOP = "ta_p_doubletop"
    DOUBLE_BOTTOM = "ta_p_doublebottom"
    MULTIPLE_TOP = "ta_p_multipletop"
    MULTIPLE_BOTTOM = "ta_p_multiplebottom"
    HEAD_AND_SHOULDERS = "ta_p_headandshoulders"
    HEAD_AND_SHOULDERS_INVERSE = "ta_p_headandshouldersinv"


class Table(EnumWithValues):
    OVERVIEW = "111"



import string
import typing as ty

from finavis.library.types import Decimal


def text_to_label(value: str) -> ty.Optional[str]:
    """Some `Hello world` convert to `hello_world`"""

    result: ty.List[str] = list()

    keys: ty.Set[str] = set(list(string.ascii_lowercase) + list(string.digits))
    for key in value.strip().lower():
        if key == "%":
            result.append("percent")
            continue
        if key in keys:
            result.append(key)
        else:
            if len(result) == 0:
                result.append("_")
            elif "_" != result[-1]:
                result.append("_")

    if len(result) == 0:
        return None
    else:
        line: str = "".join(result)

    if line.startswith("_"):
        line = line[1 : len(line)]

    if line.endswith("_"):
        line = line[0 : len(line) - 1]

    if line.startswith("52w"):
        line = line.replace("52w", "ttm")

    return line


def text_to_decimal(value: str) -> ty.Optional[Decimal]:
    """Helper func"""

    if value is None:
        return None

    mapp = {
        "M": Decimal("1000000"),
        "B": Decimal("1000000000"),
    }

    sym: str = value[-1]
    return Decimal(str(int(Decimal(value.split(sym)[0]) * mapp[sym])))



import functools
import typing as ty

from finavis.library import Quote
from finavis.utils import make_request, text_to_label

if ty.TYPE_CHECKING:
    from lxml.html import HtmlElement


@functools.lru_cache(maxsize=32, typed=False)
def get_quote(ticker: str) -> Quote:
    """Receive info by ticker name"""

    if not isinstance(ticker, str):
        raise TypeError(
            f"Argument `ticker` is not a string (currently: type({type(ticker)}))"
        )

    data: ty.Dict[str, ty.Any] = dict()
    raw: "HtmlElement" = make_request(
        path="/quote.ashx",
        query_params={"t": str(ticker)},
    )
    title: "HtmlElement" = raw.cssselect('div[class="fv-container py-2.5"]')[0]

    ticker = title.cssselect(
        'h1[class="js-recent-quote-ticker quote-header_ticker-wrapper_ticker"]'
    )[0].text_content()
    data["ticker"] = str(ticker).strip()

    try:
        company = title.cssselect(
            'h2[class="quote-header_ticker-wrapper_company text-xl"]'
        )[0]
    except IndexError:
        company = title.cssselect('h2[class="quote-header_ticker-wrapper_company"]')[0]

    data["company"] = company.text_content().strip()

    website = company.cssselect('a[class="tab-link block truncate"]')[0].attrib["href"]
    if str(website).startswith("http"):
        data["website"] = website

    keys = ["Sector", "Industry", "Country", "Exchange"]
    fields = [f.text_content() for f in title.cssselect('a[class="tab-link"]')]
    data_raw = dict(zip(keys, fields))
    for key, value in data_raw.items():
        data[key.lower()] = value

    for rows in raw.cssselect('tr[class="table-dark-row"]'):
        rows = rows.xpath("td")
        for index in range(0, 11, 2):
            field_label, field_data = rows[index], rows[index + 1]

            name: str = field_label.text_content().strip()
            label: str = ty.cast(str, text_to_label(value=name))
            if label in data.keys() and name == "EPS next Y":
                name = "EPS growth next Y"
                label = ty.cast(str, text_to_label(value=name))

            value = field_data.text_content().strip()  # type: ignore[assignment]
            if value == "-":
                value = None  # type: ignore[assignment]

            data[label] = value

    return Quote.from_response(raw=data)


def get_quotes(tickers: ty.Tuple[str, ...]) -> ty.Tuple[Quote]:
    """Receive info by tickers"""

    result: ty.Any = list()
    for ticker in tickers:
        result.append(get_quote(ticker=ticker))

    return tuple(result)



from __future__ import annotations

import datetime as dt
import typing as ty
from copy import deepcopy

import attr

from finavis.library.types import Decimal
from finavis.utils import text_to_decimal


class AbstractModel:
    def to_dict(self) -> ty.Dict[str, str]:
        return deepcopy(self.__dict__)

    @classmethod
    def from_response(cls, raw: ty.Dict[str, ty.Any]) -> ty.Any:
        raise NotImplemented

    @property
    def is_filled(self) -> bool:
        return bool(list(filter(None, self.to_dict().values())))


@attr.s(auto_attribs=True)
class Quote(AbstractModel):
    """
    Model of quote.

    :param str ticker: Ticker name
    :param str company: Company name
    :param str website: Company website
    :param str sector: Name of sector
    :param str industry: Name of industry
    :param str country: Country
    :param str exchange: Exchange
    :param ty.Tuple[str, ...] index: Major index membership
    :param Decimal p_e: Price-to-Earnings (ttm)
    :param Decimal eps_ttm: Diluted EPS (ttm)
    :param Decimal insider_own: Insider ownership
    :param Decimal shs_outstand: Shares outstanding
    :param Decimal perf_week: Performance (Week)
    :param Decimal market_cap: Market capitalization
    :param Decimal forward_p_e: Forward Price-to-Earnings (next fiscal year)
    :param Decimal eps_next_y: EPS estimate for next year
    :param Decimal insider_trans: Insider transactions (6-Month change in Insider Ownership)
    :param Decimal shs_float: Shares float
    :param Decimal perf_month: Performance (Month)
    :param Decimal income: Income (ttm)
    :param Decimal peg: Price-to-Earnings-to-Growth
    :param Decimal eps_next_q: EPS estimate for next quarter
    :param Decimal inst_own: Institutional ownership
    :param str short_float_ratio: Short interest share / ratio
    :param Decimal perf_quarter: Performance (Quarter)
    :param Decimal sales: Revenue (ttm)
    :param Decimal p_s: Price-to-Sales (ttm)
    :param Decimal eps_this_y: EPS growth this year
    :param Decimal inst_trans: Institutional transactions (3-Month change in Institutional Ownership)
    :param Decimal short_interest: Short interest
    :param Decimal perf_half_y: Performance (Half Year)
    :param Decimal book_sh: Book value per share (mrq)
    :param Decimal p_b: Price-to-Book (mrq)
    :param Decimal eps_growth_next_y: EPS growth next year
    :param Decimal roa: Return on Assets (ttm)
    :param Decimal target_price: Analysts' mean target price
    :param Decimal perf_year: Performance (Year)
    :param Decimal cash_sh: Cash per share (mrq)
    :param Decimal p_c: Price to cash per share (mrq)
    :param Decimal eps_next_5y: Long term annual growth estimate (5 years)
    :param Decimal roe: Return on Equity (ttm)
    :param str ttm_range: 52-Week trading range
    :param Decimal perf_ytd: Performance (Year To Date)
    :param Decimal dividend: Dividend (Fiscal Year Estimate)
    :param Decimal p_fcf: Price to Free Cash Flow (ttm)
    :param Decimal eps_past_5y: Annual EPS growth past 5 years
    :param Decimal roi: Return on Investment (ttm)
    :param Decimal ttm_high: Distance from 52-Week High
    :param Decimal beta: Beta
    :param Decimal dividend_percent: Dividend yield (Fiscal Year Estimate)
    :param Decimal quick_ratio: Quick Ratio (mrq)
    :param Decimal sales_past_5y: Annual sales growth past 5 years
    :param Decimal gross_margin: Gross Margin (ttm)
    :param Decimal ttm_low: Distance from 52-Week Low
    :param Decimal atr: Average True Range (14)
    :param Decimal employees: Full time employees
    :param Decimal current_ratio: Current Ratio (mrq)
    :param Decimal sales_q_q: Quarterly revenue growth (YoY)
    :param Decimal oper_margin: Operating Margin (ttm)
    :param Decimal rsi_14: Relative Strength Index
    :param str volatility_w: Volatility in week
    :param str volatility_m: Volatility in month
    :param bool is_optionable: Stock has options trading on a market exchange
    :param Decimal debt_eq: Total Debt to Equity (mrq)
    :param Decimal eps_q_q: Quarterly earnings growth (YoY)
    :param Decimal profit_margin: Net Profit Margin (ttm)
    :param Decimal rel_volume: Relative volume
    :param Decimal prev_close: Previous close
    :param bool is_shortable: Stock available to sell short
    :param Decimal lt_debt_eq: Long Term Debt to Equity (mrq)
    :param dt.date earnings_at: Earnings date
    :param str earnings_market: BMO = Before Market Open, AMC = After Market Close
    :param Decimal payout: Dividend Payout Ratio (ttm)
    :param Decimal avg_volume: Average volume (3 month)
    :param Decimal price: Current stock price
    :param Decimal recom: Analysts' mean recommendation (1=Buy 5=Sell)
    :param Decimal sma20: Distance from 20-Day Simple Moving Average
    :param Decimal sma50: Distance from 50-Day Simple Moving Average
    :param Decimal sma200: Distance from 200-Day Simple Moving Average
    :param Decimal volume: Volume
    :param Decimal change: Performance (today)
    """

    ticker: str
    company: str
    website: str
    sector: str
    industry: str
    country: str
    exchange: str
    p_e: Decimal
    eps_ttm: Decimal
    insider_own: Decimal
    shs_outstand: Decimal
    perf_week: Decimal
    market_cap: Decimal
    forward_p_e: Decimal
    eps_next_y: Decimal
    insider_trans: Decimal
    shs_float: Decimal
    perf_month: Decimal
    income: Decimal
    peg: Decimal
    eps_next_q: Decimal
    inst_own: Decimal
    short_float_ratio: str
    perf_quarter: Decimal
    sales: Decimal
    p_s: Decimal
    eps_this_y: Decimal
    inst_trans: Decimal
    short_interest: Decimal
    perf_half_y: Decimal
    book_sh: Decimal
    p_b: Decimal
    eps_growth_next_y: Decimal
    roa: Decimal
    target_price: Decimal
    perf_year: Decimal
    cash_sh: Decimal
    p_c: Decimal
    eps_next_5y: Decimal
    roe: Decimal
    ttm_range: str
    perf_ytd: Decimal
    dividend: Decimal
    p_fcf: Decimal
    eps_past_5y: Decimal
    roi: Decimal
    ttm_high: Decimal
    beta: Decimal
    dividend_percent: Decimal
    quick_ratio: Decimal
    sales_past_5y: Decimal
    gross_margin: Decimal
    ttm_low: Decimal
    atr: Decimal
    employees: Decimal
    current_ratio: Decimal
    sales_q_q: Decimal
    oper_margin: Decimal
    rsi_14: Decimal
    volatility_w: Decimal
    volatility_m: Decimal
    debt_eq: Decimal
    eps_q_q: Decimal
    profit_margin: Decimal
    rel_volume: Decimal
    prev_close: Decimal
    lt_debt_eq: Decimal
    earnings_market: str
    payout: Decimal
    avg_volume: Decimal
    price: Decimal
    recom: Decimal
    sma20: Decimal
    sma50: Decimal
    sma200: Decimal
    volume: Decimal
    change: Decimal
    is_optionable: ty.Optional[bool] = None
    is_shortable: ty.Optional[bool] = None
    index: ty.Optional[ty.Tuple[str, ...]] = None
    earnings_at: ty.Optional[dt.date] = None

    def to_dict(self) -> ty.Dict[str, str]:
        raw: ty.Dict[str, str] = super().to_dict()
        for key, value in self.__dict__.items():
            if key.startswith("ttm_"):
                raw["52w_" + key.split("ttm_")[-1]] = value
        return raw

    @classmethod
    def from_response(cls, raw: ty.Dict[str, ty.Any]) -> Quote:
        """Make class from raw_data"""

        data: ty.Dict[str, ty.Any] = dict()

        index = raw.pop("index", None)
        if isinstance(index, str):
            data.update(index=tuple((x.strip() for x in index.split(","))))

        optionable = raw.pop("optionable", None)
        if optionable is not None:
            data.update(is_optionable=optionable == "Yes")

        shortable = raw.pop("shortable", None)
        if shortable is not None:
            data.update(is_shortable=shortable == "Yes")

        earnings = raw.pop("earnings", None)
        if isinstance(earnings, str):
            data.update(earnings_market=earnings.split()[-1])

            today = dt.date.today()
            earn_at = dt.datetime.strptime(
                " ".join([str(today.year)] + earnings.split()[0:2]), "%Y %b %d"
            ).date()
            if earn_at < today:
                earn_at = dt.date(today.year + 1, earn_at.month, earn_at.day)

            data.update(earnings_at=earn_at)

        volatility = raw.pop("volatility", None)
        if isinstance(volatility, str):
            volatility_w, volatility_m = volatility.split()
            if isinstance(volatility_w, str):
                data.update(volatility_w=Decimal(volatility_w.split("%")[0]))
            if isinstance(volatility_m, str):
                data.update(volatility_m=Decimal(volatility_m.split("%")[0]))

        for field_name, field_type in cls.__annotations__.items():
            value: ty.Any = raw.get(field_name)

            if field_name in data.keys():
                continue

            if field_type == "str":
                data[field_name] = value
                continue

            if field_type == "Decimal":
                if value is None:
                    data[field_name] = value
                    continue

                if value.endswith(("M", "B")):
                    data[field_name] = text_to_decimal(value=value)
                    continue

                if value.endswith("%"):
                    data[field_name] = Decimal(value.split("%")[0])
                    continue

                if "," in value:
                    value = value.replace(",", "")

                data[field_name] = Decimal(value)

        return cls(**data)


@attr.s(auto_attribs=True)
class Overview(AbstractModel):
    """
    Model of overview table in screener.

    :param str ticker: Ticker
    :param str company: Company
    :param str sector: Sector
    :param str industry: Industry
    :param str country: Country
    :param Decimal market_cap: Market capitalization
    :param Decimal p_e: Price-to-Earnings (ttm)
    :param Decimal price: Current stock price
    :param Decimal change: Performance (today)
    :param Decimal volume: Volume
    """

    ticker: str
    company: str
    sector: str
    industry: str
    country: str
    market_cap: Decimal
    p_e: Decimal
    price: Decimal
    change: Decimal
    volume: Decimal

    @classmethod
    def from_response(cls, raw: ty.Dict[str, ty.Any]) -> Overview:
        """Make class from raw_data"""

        data: ty.Dict[str, ty.Any] = dict()
        for field_name, field_type in cls.__annotations__.items():
            value: ty.Any = raw.get(field_name)

            if field_type == "Decimal":
                if value is None:
                    data[field_name] = value
                    continue

                if value.endswith("%"):
                    data[field_name] = Decimal(value.split("%")[0])
                    continue

                if value.endswith(("M", "B")):
                    data[field_name] = text_to_decimal(value=value)
                    continue

                if "," in value:
                    value = value.replace(",", "")

                data[field_name] = Decimal(value)
            else:
                data[field_name] = value

        return Overview(**data)



class RequestUnhandledException(Exception):
    """Unhandled"""


class NotFoundException(Exception):
    """Custom exception if data not found"""


class TickerNotFoundException(Exception):
    """Custom exception if ticker not found"""


class RequestTimeoutException(Exception):
    """Exception if timeout"""


class RequestMaxRetryException(Exception):
    """Exception if max_retries on session"""


class RequestDocumentIsEmptyException(Exception):
    """If document is empty"""



import re
import typing as ty

from finavis.library import Exchange, Index, Order, Overview, Signal, Table
from finavis.utils.sessions import make_request

if ty.TYPE_CHECKING:
    from lxml import html


class Screener:
    """Screen! Screen! Screener!"""

    _args_validate_mapping: ty.Dict[str, ty.Any] = dict(
        exchange=Exchange,
        index=Index,
        signal=Signal,
        table=Table,
        order_by=Order,
    )
    _screener_mapping: ty.Dict[str, ty.Any] = {Table.OVERVIEW.value: Overview}
    _per_pages: int = 20
    _items: ty.List[Overview] = list()

    def __len__(self) -> int:
        """Total objects in array"""

        return len(self._items)

    def __getitem__(self, index):
        """Getting objects by index"""

        return self._items[index]

    def __next__(self):
        for item in self._items:
            yield item

        raise StopIteration

    get = __getitem__

    def __init__(
        self,
        exchange: ty.Optional[ty.Union[Exchange, str]] = None,
        index: ty.Optional[ty.Union[Index, str]] = None,
        signal: ty.Optional[ty.Union[Signal, str]] = None,
        table: ty.Optional[ty.Union[Table, str]] = Table.OVERVIEW,
        order_by: ty.Union[Order, str] = Order.TICKER_ASC,
    ) -> None:
        """Initialization and validation"""

        self.exchange = str(exchange) if exchange is not None else None
        self.index = str(index) if index is not None else None
        self.signal = str(signal) if signal is not None else None
        self.table = str(table) if table is not None else None
        self.order_by = str(order_by) if order_by is not None else None

        for arg_name, arg_choices in self._args_validate_mapping.items():
            value: ty.Optional[str] = getattr(self, arg_name)
            if value is not None and str(value) not in arg_choices:
                raise TypeError(
                    f"arg {arg_name}={value} is not allowed, please select "
                    f"some another, if required: {', '.join(arg_choices.values())}."
                )

        self.filters: ty.List[str] = list(filter(None, [self.exchange, self.index]))

        self.total: int = 0
        self.pages: int = 0

    def __repr__(self) -> str:
        """String representation of class"""

        return (
            f"<{self.__class__.__name__} "
            f"exchange={self.exchange}, index={self.index}, signal={self.signal}, "
            f"table={self.table}, order_by={self.order_by}, total={self.total}, "
            f"pages={self.pages}>"
        )

    def __call__(self) -> ty.List[Overview]:
        if len(self._items) > 0:
            return self._items

        for index, item in enumerate(self._yielding_objects(page=1)):
            self._items.append(item)

        return self._items

    def _yielding_objects(self, page: int = 1) -> ty.Iterable[ty.Union[Overview]]:
        """Getting data and create object"""

        query_params: ty.Dict[str, ty.Any] = dict(
            v=self.table,
            o=self.order_by,
            r=((page * self._per_pages) - self._per_pages) + 1,
        )

        if len(self.filters) > 0:
            query_params.update(f=",".join(self.filters))

        raw: "html.HtmlElement" = make_request(
            path="/screener.ashx",
            query_params=query_params,
        )

        if not self.total:
            try:
                total_raw = raw.get_element_by_id("screener-total")
                total_raw = total_raw.text_content()  # type: ignore[assignment]
            except IndexError:
                return None
            else:
                total_raw_lines = re.findall(r"\s\d+", str(total_raw))
                if len(total_raw_lines) > 0:
                    self.total = int(total_raw_lines[0].strip())
                    self.pages = self._get_total_pages(
                        total=self.total / self._per_pages
                    )

        if not self.total:
            return None

        try:
            raw = raw.get_element_by_id("screener-table")
        except KeyError:
            return None

        for raw_item in raw.cssselect("tr")[3:]:
            key: str = self._screener_mapping[self.table].__name__.lower()  # type: ignore[index]
            yield getattr(self, f"_get_{key}")(raw=raw_item)

        if page < self.pages:
            yield from self._yielding_objects(page=page + 1)

    @staticmethod
    def _get_overview(raw: "html.HtmlElement") -> Overview:
        """Make overview object"""

        kwargs: ty.Dict[str, str] = dict(
            zip(
                Overview.__annotations__.keys(),
                [x if x != "-" else None for x in raw.xpath("td//text()")[1:]],
            )
        )

        return Overview.from_response(raw=kwargs)

    @staticmethod
    def _get_total_pages(total: float) -> int:
        """Helper func"""

        pages, tail = str(total).split(".")
        return int(pages) + 1 if int(tail) > 0 else int(pages)
